from collections import deque
import random

class ReplayBuffer(object):

    def __init__(self, buffer_size):
        self.buffer_size = buffer_size
        self.num_experiences = 0
        self.buffer = deque()

    def getBatch(self, batch_size):
        # Randomly sample batch_size examples
        if self.num_experiences < batch_size:
            return random.sample(self.buffer, self.num_experiences)
        else:
            return random.sample(self.buffer, batch_size)

    def size(self):
        return self.buffer_size

    def add(self, state, action, reward, new_state, done):
        experience = (state, action, reward, new_state, done)
        if self.num_experiences < self.buffer_size:
            self.buffer.append(experience)
            self.num_experiences += 1
        else:
            self.buffer.popleft()
            self.buffer.append(experience)




import pygame
import sys
def load():
    # path of player with different states
    import os
    dir_path = os.path.dirname(os.path.realpath(__file__))
    PLAYER_PATH = (
            dir_path+'/assets/sprites/redbird-upflap.png',
            dir_path+'/assets/sprites/redbird-midflap.png',
            dir_path+'/assets/sprites/redbird-downflap.png'
    )



    # path of background
    BACKGROUND_PATH = dir_path+'/assets/sprites/background-black.png'

    # path of pipe
    PIPE_PATH = dir_path+'/assets/sprites/pipe-green.png'

    IMAGES, SOUNDS, HITMASKS = {}, {}, {}

    # numbers sprites for score display
    IMAGES['numbers'] = (
        pygame.image.load(dir_path+'/assets/sprites/0.png').convert_alpha(),
        pygame.image.load(dir_path+'/assets/sprites/1.png').convert_alpha(),
        pygame.image.load(dir_path+'/assets/sprites/2.png').convert_alpha(),
        pygame.image.load(dir_path+'/assets/sprites/3.png').convert_alpha(),
        pygame.image.load(dir_path+'/assets/sprites/4.png').convert_alpha(),
        pygame.image.load(dir_path+'/assets/sprites/5.png').convert_alpha(),
        pygame.image.load(dir_path+'/assets/sprites/6.png').convert_alpha(),
        pygame.image.load(dir_path+'/assets/sprites/7.png').convert_alpha(),
        pygame.image.load(dir_path+'/assets/sprites/8.png').convert_alpha(),
        pygame.image.load(dir_path+'/assets/sprites/9.png').convert_alpha()
    )

    # base (ground) sprite
    IMAGES['base'] = pygame.image.load(dir_path+'/assets/sprites/base.png').convert_alpha()

    # sounds
    if 'win' in sys.platform:
        soundExt = '.wav'
    else:
        soundExt = '.ogg'

    SOUNDS['die']    = pygame.mixer.Sound(dir_path+'/assets/audio/die' + soundExt)
    SOUNDS['hit']    = pygame.mixer.Sound(dir_path+'/assets/audio/hit' + soundExt)
    SOUNDS['point']  = pygame.mixer.Sound(dir_path+'/assets/audio/point' + soundExt)
    SOUNDS['swoosh'] = pygame.mixer.Sound(dir_path+'/assets/audio/swoosh' + soundExt)
    SOUNDS['wing']   = pygame.mixer.Sound(dir_path+'/assets/audio/wing' + soundExt)

    # select random background sprites
    IMAGES['background'] = pygame.image.load(BACKGROUND_PATH).convert()

    # select random player sprites
    IMAGES['player'] = (
        pygame.image.load(PLAYER_PATH[0]).convert_alpha(),
        pygame.image.load(PLAYER_PATH[1]).convert_alpha(),
        pygame.image.load(PLAYER_PATH[2]).convert_alpha(),
    )

    # select random pipe sprites
    IMAGES['pipe'] = (
        pygame.transform.rotate(
            pygame.image.load(PIPE_PATH).convert_alpha(), 180),
        pygame.image.load(PIPE_PATH).convert_alpha(),
    )

    # hismask for pipes
    HITMASKS['pipe'] = (
        getHitmask(IMAGES['pipe'][0]),
        getHitmask(IMAGES['pipe'][1]),
    )

    # hitmask for player
    HITMASKS['player'] = (
        getHitmask(IMAGES['player'][0]),
        getHitmask(IMAGES['player'][1]),
        getHitmask(IMAGES['player'][2]),
    )

    return IMAGES, SOUNDS, HITMASKS

def getHitmask(image):
    """returns a hitmask using an image's alpha."""
    mask = []
    for x in range(image.get_width()):
        mask.append([])
        for y in range(image.get_height()):
            mask[x].append(bool(image.get_at((x,y))[3]))
    return mask



import random
from itertools import cycle

import pygame

import flappy_bird_utils

FPS = 30
SCREENWIDTH  = 288
SCREENHEIGHT = 512

pygame.init()
FPSCLOCK = pygame.time.Clock()
SCREEN = pygame.display.set_mode((SCREENWIDTH, SCREENHEIGHT))
pygame.display.set_caption('Flappy Bird')

IMAGES, SOUNDS, HITMASKS = flappy_bird_utils.load()
PIPEGAPSIZE = 100 # gap between upper and lower part of pipe
BASEY = SCREENHEIGHT * 0.79

PLAYER_WIDTH = IMAGES['player'][0].get_width()
PLAYER_HEIGHT = IMAGES['player'][0].get_height()
PIPE_WIDTH = IMAGES['pipe'][0].get_width()
PIPE_HEIGHT = IMAGES['pipe'][0].get_height()
BACKGROUND_WIDTH = IMAGES['background'].get_width()

PLAYER_INDEX_GEN = cycle([0, 1, 2, 1])


class GameState:
    def __init__(self):
        self.score = self.playerIndex = self.loopIter = 0
        self.playerx = int(SCREENWIDTH * 0.2)
        self.playery = int((SCREENHEIGHT - PLAYER_HEIGHT) / 2)
        self.basex = 0
        self.baseShift = IMAGES['base'].get_width() - BACKGROUND_WIDTH

        newPipe1 = getRandomPipe()
        newPipe2 = getRandomPipe()
        self.upperPipes = [
            {'x': SCREENWIDTH, 'y': newPipe1[0]['y']},
            {'x': SCREENWIDTH + (SCREENWIDTH / 2), 'y': newPipe2[0]['y']},
        ]
        self.lowerPipes = [
            {'x': SCREENWIDTH, 'y': newPipe1[1]['y']},
            {'x': SCREENWIDTH + (SCREENWIDTH / 2), 'y': newPipe2[1]['y']},
        ]

        # player velocity, max velocity, downward accleration, accleration on flap
        self.pipeVelX = -4
        self.playerVelY    =  0    # player's velocity along Y, default same as playerFlapped
        self.playerMaxVelY =  10   # max vel along Y, max descend speed
        self.playerMinVelY =  -8   # min vel along Y, max ascend speed
        self.playerAccY    =   1   # players downward accleration
        self.playerFlapAcc =  -9   # players speed on flapping
        self.playerFlapped = False # True when player flaps

    def frame_step(self, input_actions):
        pygame.event.pump()

        reward = 0.1
        terminal = False

        if sum(input_actions) != 1:
            raise ValueError('Multiple input actions!')

        # input_actions[0] == 1: do nothing
        # input_actions[1] == 1: flap the bird
        if input_actions[1] == 1:
            if self.playery > -2 * PLAYER_HEIGHT:
                self.playerVelY = self.playerFlapAcc
                self.playerFlapped = True
                #SOUNDS['wing'].play()

        # check for score
        playerMidPos = self.playerx + PLAYER_WIDTH / 2
        for pipe in self.upperPipes:
            pipeMidPos = pipe['x'] + PIPE_WIDTH / 2
            if pipeMidPos <= playerMidPos < pipeMidPos + 4:
                self.score += 1
                #SOUNDS['point'].play()
                reward = 1

        # playerIndex basex change
        if (self.loopIter + 1) % 3 == 0:
            self.playerIndex = next(PLAYER_INDEX_GEN)
        self.loopIter = (self.loopIter + 1) % 30
        self.basex = -((-self.basex + 100) % self.baseShift)

        # player's movement
        if self.playerVelY < self.playerMaxVelY and not self.playerFlapped:
            self.playerVelY += self.playerAccY
        if self.playerFlapped:
            self.playerFlapped = False
        self.playery += min(self.playerVelY, BASEY - self.playery - PLAYER_HEIGHT)
        if self.playery < 0:
            self.playery = 0

        # move pipes to left
        for uPipe, lPipe in zip(self.upperPipes, self.lowerPipes):
            uPipe['x'] += self.pipeVelX
            lPipe['x'] += self.pipeVelX

        # add new pipe when first pipe is about to touch left of screen
        if 0 < self.upperPipes[0]['x'] < 5:
            newPipe = getRandomPipe()
            self.upperPipes.append(newPipe[0])
            self.lowerPipes.append(newPipe[1])

        # remove first pipe if its out of the screen
        if self.upperPipes[0]['x'] < -PIPE_WIDTH:
            self.upperPipes.pop(0)
            self.lowerPipes.pop(0)

        # check if crash here
        isCrash= checkCrash({'x': self.playerx, 'y': self.playery,
                             'index': self.playerIndex},
                            self.upperPipes, self.lowerPipes)
        if isCrash:
            #SOUNDS['hit'].play()
            #SOUNDS['die'].play()
            terminal = True
            self.__init__()
            reward = -1

        # draw sprites
        SCREEN.blit(IMAGES['background'], (0,0))

        for uPipe, lPipe in zip(self.upperPipes, self.lowerPipes):
            SCREEN.blit(IMAGES['pipe'][0], (uPipe['x'], uPipe['y']))
            SCREEN.blit(IMAGES['pipe'][1], (lPipe['x'], lPipe['y']))

        SCREEN.blit(IMAGES['base'], (self.basex, BASEY))
        # print score so player overlaps the score
        # showScore(self.score)
        SCREEN.blit(IMAGES['player'][self.playerIndex],
                    (self.playerx, self.playery))

        image_data = pygame.surfarray.array3d(pygame.display.get_surface())
        pygame.display.update()
        FPSCLOCK.tick(FPS)
        #print self.upperPipes[0]['y'] + PIPE_HEIGHT - int(BASEY * 0.2)
        return image_data, reward, terminal

def getRandomPipe():
    """returns a randomly generated pipe"""
    # y of gap between upper and lower pipe
    gapYs = [20, 30, 40, 50, 60, 70, 80, 90]
    index = random.randint(0, len(gapYs)-1)
    gapY = gapYs[index]

    gapY += int(BASEY * 0.2)
    pipeX = SCREENWIDTH + 10

    return [
        {'x': pipeX, 'y': gapY - PIPE_HEIGHT},  # upper pipe
        {'x': pipeX, 'y': gapY + PIPEGAPSIZE},  # lower pipe
    ]


def showScore(score):
    """displays score in center of screen"""
    scoreDigits = [int(x) for x in list(str(score))]
    totalWidth = 0 # total width of all numbers to be printed

    for digit in scoreDigits:
        totalWidth += IMAGES['numbers'][digit].get_width()

    Xoffset = (SCREENWIDTH - totalWidth) / 2

    for digit in scoreDigits:
        SCREEN.blit(IMAGES['numbers'][digit], (Xoffset, SCREENHEIGHT * 0.1))
        Xoffset += IMAGES['numbers'][digit].get_width()


def checkCrash(player, upperPipes, lowerPipes):
    """returns True if player collders with base or pipes."""
    pi = player['index']
    player['w'] = IMAGES['player'][0].get_width()
    player['h'] = IMAGES['player'][0].get_height()

    # if player crashes into ground
    if player['y'] + player['h'] >= BASEY - 1:
        return True
    else:

        playerRect = pygame.Rect(player['x'], player['y'],
                      player['w'], player['h'])

        for uPipe, lPipe in zip(upperPipes, lowerPipes):
            # upper and lower pipe rects
            uPipeRect = pygame.Rect(uPipe['x'], uPipe['y'], PIPE_WIDTH, PIPE_HEIGHT)
            lPipeRect = pygame.Rect(lPipe['x'], lPipe['y'], PIPE_WIDTH, PIPE_HEIGHT)

            # player and upper/lower pipe hitmasks
            pHitMask = HITMASKS['player'][pi]
            uHitmask = HITMASKS['pipe'][0]
            lHitmask = HITMASKS['pipe'][1]

            # if bird collided with upipe or lpipe
            uCollide = pixelCollision(playerRect, uPipeRect, pHitMask, uHitmask)
            lCollide = pixelCollision(playerRect, lPipeRect, pHitMask, lHitmask)

            if uCollide or lCollide:
                return True

    return False

def pixelCollision(rect1, rect2, hitmask1, hitmask2):
    """Checks if two objects collide and not just their rects"""
    rect = rect1.clip(rect2)

    if rect.width == 0 or rect.height == 0:
        return False

    x1, y1 = rect.x - rect1.x, rect.y - rect1.y
    x2, y2 = rect.x - rect2.x, rect.y - rect2.y

    for x in range(rect.width):
        for y in range(rect.height):
            if hitmask1[x1+x][y1+y] and hitmask2[x2+x][y2+y]:
                return True
    return False



import gym
import cv2

class GameEngine:
    def __init__(self):
        print("game engine initialized")
    def initialState(self):
        raise NotImplementedError( "Should have implemented this, return R, S, T" )
    def step(self, action, state=None):
        raise NotImplementedError( "Should have implemented this" )

class OpenAIGameEngine(GameEngine):
    def __init__(self, ENVIRONMENT_NAME):
        self.envName = ENVIRONMENT_NAME
        self.env = gym.make(ENVIRONMENT_NAME)

    def initialState(self):
        s = self.env.reset()
        return s
    def step(self, action, state=None): # return r, s ,t
        s1, r, terminated, _ = self.env.step(action)
        return r, s1, terminated

    def gameParameters(self):
        state_size = self.env.observation_space.shape[0]
        action_size = self.env.action_space.shape[0]
        action_high = self.env.action_space.high
        action_low = self.env.action_space.low
        return state_size, action_size, action_high, action_low
    def show(self):
        self.env.render()

    def monitor(self, task):
        self.env.monitor.start('experiments/' + self.envName,force=True)
        task()
        self.env.monitor.close()



class FlappyBirdGameEngine(GameEngine):
    def __init__(self):
        pass

    def initialState(self):
        import FlappyBirdEnv.wrapped_flappy_bird as game
        game_state = game.GameState()
        return game_state
    def step(self, action, state=None): # return r, s ,t
        ORIGIN_LENG = 160
        ORIGIN_WIDTH = 160
        x_t1_colored, r_t, terminal = state.frame_step(action)
        x_t1 = cv2.cvtColor(cv2.resize(x_t1_colored, (ORIGIN_LENG, ORIGIN_WIDTH)), cv2.COLOR_BGR2GRAY)
        ret, x_t1 = cv2.threshold(x_t1, 1, 255, cv2.THRESH_BINARY)
        return ret, x_t1, terminal







import tensorflow as tf
from NN import NN
class Actor(NN):
    # variables into a bag
    # ema wrapper
    # gradient wrapper
    def __init__(self, session, hasShadowNet, state_size, action_size, hidden_state_size):
        NN.__init__(self, session, hasShadowNet)
        # nothing special:
        inputLayer = self.buildInputLayer("inputStates", shape=[None, state_size])
        h1 = self.buildLinearReluWire(inputLayer, [state_size, hidden_state_size])
        #for i in range(numOfHiddenLayers-1): # repeat (numOfHiddenLayers-1) times
        h1 = self.buildLinearReluWire(h1, [hidden_state_size, hidden_state_size])
        out = self.buildLinearWire(h1, [hidden_state_size, action_size])
        self.setOutLayer(out)

        # a bit unique:
        Qgradient = self.buildInputLayer("Qgradients", shape=[None, action_size])
        self.addAscentOperation(Qgradient)



import tensorflow as tf
import sys
class NN:
    # design philosophy
    #  (1) do not let me explicitly reason about the shadow network, handle it automatically
    #  (2) do not let me do the gradient arithmetic operations.
    # (3)  do not let me remember which tensorflow variables are associated with my input data. I only remember the name



    def __init__(self, session, hasShadowNet=True): #ema = tf.train.ExponentialMovingAverage(0.999)

        self._session = session
        if hasShadowNet == True:
            self._ema = tf.train.ExponentialMovingAverage(0.999)
        else:
            self._ema = None


        self._weight_counter = 0
        self._bias_counter = 0
        self._layer_counter = 0
        self._operation_counter = 0

        self._anyNamedOperations = {}


        self._layers = {}
        self._variables = {}
        self._placeholders = {}

        self._layers2shadow={}
        self._variables2shadow = {}
        self._placeholders2shadow = {}
        self._flushShadowNetOps = []


        self._graphviz = ""






    def weight_variable(self, shape):
        initial = tf.truncated_normal(shape, stddev = 0.01)
        return tf.Variable(initial)

    def bias_variable(self, shape):
        initial = tf.constant(0.01, shape=shape)
        return tf.Variable(initial)



    def uniqueWeightName(self):
        self._weight_counter+=1
        return "W" + str(self._weight_counter)

    def uniqueBiasName(self):
        self._bias_counter+=1
        return "B" + str(self._bias_counter)

    def uniqueLayerName(self):
        self._layer_counter+=1
        return "L" + str(self._layer_counter)



    def shape2string(self, weightShape):
        string = ""
        index = 0
        for i in weightShape:
            dimString = str(i)
            if dimString == '?':
                string += "UNK"
            else:
                string += dimString

            if index != len(weightShape)-1:
                string += "*"
            index += 1
        return string

    def shadowPrefixed(self, x):
        return "Shadow"+x



    def buildInputLayer(self, name, shape): # we assign a name because we will access it later
        holder = tf.placeholder("float", shape=shape)
        self._layers[name] = holder
        self._placeholders[name] = holder

        self._layers2shadow[holder] = holder # reuse the inputs in both the basic net and the shadow net
        self._placeholders2shadow[holder] = holder # shared input

        return holder


    def buildLinearReluWire(self, layer, weightShape):
        # basic
        W1 = self.weight_variable(shape = weightShape)
        B1 = self.bias_variable(shape=[weightShape[-1]])
        tmp = tf.matmul(layer, W1) + B1
        outputLayer = tf.nn.relu(tmp)
        weightName = self.uniqueWeightName()
        self._variables[weightName] = W1 # register the created variables
        biasName = self.uniqueBiasName()
        self._variables[biasName] = B1
        layerName = self.uniqueLayerName()
        self._layers[layerName] = outputLayer # register the layer

        # visualization
        label = weightName+":"+ self.shape2string(weightShape) + "," + biasName + ":"+self.shape2string([weightShape[-1]])
        self.createGraphvizEdge(self.lookupLayerName(layer), layerName, label)

        # shadow part:
        if self._ema != None:
            flush = self._ema.apply([W1, B1]) # create some symbolic representation of shadow, otherwise, the following "average" call would return None
            self._flushShadowNetOps.append(flush)
            W1shadow = self._ema.average(W1)
            B1shadow = self._ema.average(B1)
            layerShadow = self._layers2shadow[layer]
            tmpshadow = tf.matmul(layerShadow, W1shadow) + B1shadow
            outputLayershadow = tf.nn.relu(tmpshadow)
            self._variables2shadow[W1] = W1shadow
            self._variables2shadow[B1] = B1shadow
            self._layers2shadow[outputLayer] = outputLayershadow
            self.createGraphvizEdge(self.shadowPrefixed(self.lookupLayerName(layer)), self.shadowPrefixed(layerName), label)


        return outputLayer

    def buildJointLinearReluWire(self, layer, weightShape, layer2, weightShape2):
        # basic
        W1 = self.weight_variable(shape = weightShape)
        B = self.bias_variable(shape=[weightShape[-1]])
        W2 = self.weight_variable(shape = weightShape2)
        if(weightShape[-1] != weightShape2[-1]):
            print("they should agree!")
            sys.exit(1)


        tmp = tf.matmul(layer, W1) + tf.matmul(layer2, W2)  + B
        outputLayer = tf.nn.relu(tmp)
        weightName = self.uniqueWeightName()
        self._variables[weightName] = W1 # register the created variables
        weightName2 = self.uniqueWeightName()
        self._variables[weightName2] = W2 # register the created variables
        biasName = self.uniqueBiasName()
        self._variables[biasName] = B
        layerName = self.uniqueLayerName()
        self._layers[layerName] = outputLayer # register the layer

        # visualization
        label = weightName+":"+ self.shape2string(weightShape) + "," + biasName + ":"+self.shape2string([weightShape[-1]])
        self.createGraphvizEdge(self.lookupLayerName(layer), layerName, label)

        label2 = weightName2+":"+ self.shape2string(weightShape2) + "," + biasName + ":"+self.shape2string([weightShape2[-1]])
        self.createGraphvizEdge(self.lookupLayerName(layer2), layerName, label2)

        # shadow part:
        if self._ema != None:
            flush = self._ema.apply([W1, W2, B]) # create some symbolic representation of shadow, otherwise, the following "average" call would return None
            self._flushShadowNetOps.append(flush)
            W1shadow = self._ema.average(W1)
            W2shadow = self._ema.average(W2)
            Bshadow = self._ema.average(B)
            layerShadow = self._layers2shadow[layer]
            layer2Shadow = self._layers2shadow[layer2]

            tmpshadow = tf.matmul(layerShadow, W1shadow) + tf.matmul(layer2Shadow, W2shadow) + Bshadow
            outputLayershadow = tf.nn.relu(tmpshadow)
            self._variables2shadow[W1] = W1shadow
            self._variables2shadow[B] = Bshadow
            self._layers2shadow[outputLayer] = outputLayershadow
            self.createGraphvizEdge(self.shadowPrefixed(self.lookupLayerName(layer)), self.shadowPrefixed(layerName), label)
            self.createGraphvizEdge(self.shadowPrefixed(self.lookupLayerName(layer2)), self.shadowPrefixed(layerName), label2)


        return outputLayer



    def buildLinearWire(self, layer, weightShape):
        W1 = self.weight_variable(shape = weightShape)
        B1 = self.bias_variable(shape=[weightShape[-1]])
        outputLayer = tf.matmul(layer, W1) + B1

        weightName = self.uniqueWeightName()
        self._variables[weightName] = W1 # register the created variables
        biasName = self.uniqueBiasName()
        self._variables[biasName] = B1
        layerName = self.uniqueLayerName()
        self._layers[layerName] = outputLayer # register the layer

        label = weightName+":"+ self.shape2string(weightShape) + "," + biasName + ":"+self.shape2string([weightShape[-1]])
        self.createGraphvizEdge(self.lookupLayerName(layer), layerName, label)

        # shadow part:
        if self._ema != None:
            flush = self._ema.apply([W1, B1]) # create some symbolic representation of shadow, otherwise, the following call would return None
            self._flushShadowNetOps.append(flush)


            W1shadow = self._ema.average(W1)
            B1shadow = self._ema.average(B1)
            layerShadow = self._layers2shadow[layer]
            outputLayershadow = tf.matmul(layerShadow, W1shadow) + B1shadow
            self._variables2shadow[W1] = W1shadow
            self._variables2shadow[B1] = B1shadow
            self._layers2shadow[outputLayer] = outputLayershadow
            self.createGraphvizEdge(self.shadowPrefixed(self.lookupLayerName(layer)), self.shadowPrefixed(layerName), label)



        return outputLayer



    def buildReduceSum(self, layer, reduction_indices):
        outputLayer = tf.reduce_sum(layer, reduction_indices = reduction_indices)
        layerName = self.uniqueLayerName()
        self._layers[layerName] = outputLayer # register the layer

        if self._ema != None:
            layerShadow = self._layers2shadow[layer]
            outputLayerShadow = tf.reduce_sum(layerShadow, reduction_indices = reduction_indices)
            self._layers2shadow[outputLayer] = outputLayerShadow

        return outputLayer









    def addDescentOperation(self, extraGradients):
        varList = self._variables.values()
        gradient = tf.gradients(self._out, varList, extraGradients)
        zipped = zip(gradient, varList)
        descent = tf.train.AdamOptimizer(1e-6).apply_gradients(zipped)
        self._descentOp = descent

        return descent

    def addAscentOperation(self, extraGradients):
        varList = self._variables.values()
        gradient = tf.gradients(self._out, varList, extraGradients)
        def multbyNegativeOne(elem):
             return elem * -1
        negativegradient = map(multbyNegativeOne, gradient)
        zipped = zip(negativegradient, varList)
        ascent = tf.train.AdamOptimizer(1e-6).apply_gradients(zipped)
        self._ascentOp = ascent
        return ascent

    def addMinimizeOperation(self, minimizer):
        self._minimize = minimizer







    def addAnyNamedOperation(self, name, op):
        self._anyNamedOperations[name] = op


    def setOutLayer(self, out):
        self._out = out
        if self._ema != None:
            self._shadowOut = self._layers2shadow[out]




    def getInputLayer(self, name):
        return self._layers[name] #


    def lookupLayerName(self, layer):
        for key, value in self._layers.iteritems():
            if value == layer:
                return key

    def createGraphvizEdge(self, name1, name2, label):
        self._graphviz += name1
        self._graphviz += " -> "
        self._graphviz += name2
        self._graphviz += " [label=\""
        self._graphviz += label
        self._graphviz += "\"]\n"

    def visualizeGraphviz(self):
        for layerName, value in self._layers.iteritems():
            if value != self._out and value not in self._placeholders.values():
                self._graphviz = layerName + "\n" +  self._graphviz
                self._graphviz = self.shadowPrefixed(layerName) + "\n" +  self._graphviz

            elif value == self._out:
                self._graphviz = layerName + " [label=Out]\n" +  self._graphviz
                self._graphviz = self.shadowPrefixed(layerName) + " [label=ShadowOut]\n" +  self._graphviz

            elif value in self._placeholders.values():
                self._graphviz = layerName + " [label=\""+layerName + ":"+ self.shape2string(value.get_shape())+ "\"]\n" +  self._graphviz
                self._graphviz = self.shadowPrefixed(layerName) + " [label=\""+ self.shadowPrefixed(layerName) + ":"+ self.shape2string(value.get_shape())+ "\"]\n" +  self._graphviz

        self._graphviz = """digraph G{
                            edge [dir=forward]
                            node [shape=plaintext]
                            """  + self._graphviz + """
                            }"""

        from graphviz import Source
        s = Source(self._graphviz, filename="test.gv", format="png")
        s.view()


    # key part: tensor flows!!!!!
    # we support three default actions: descent, ascent, minimize,
    # for any other action, we simply add it into a dictionary that holds the user-defined actions.
    def descentSmartly(self, feed_dict_string_key):
        if self._descentOp == None:
            print("add descent operation first!!!")
            sys.exit(1)

        feed_dict = {}
        for key, value in feed_dict_string_key.iteritems():
            feed_dict[self._placeholders[key]] =value

        self._session.run(self._descentOp, feed_dict=feed_dict)
        for flush in self._flushShadowNetOps:
            self._session.run(flush)

    # tensor flows!
    def ascentSmartly(self, feed_dict_string_key): # we automatically flush the updates to shadow, which is where the smart happens
        if self._ascentOp == None:
            print("add ascent operation first!!!")
            sys.exit(1)

        feed_dict = {}
        for key, value in feed_dict_string_key.iteritems():
            feed_dict[self._placeholders[key]] =value

        self._session.run(self._ascentOp, feed_dict=feed_dict)
        for flush in self._flushShadowNetOps:
            self._session.run(flush)

    def minimizeSmartly(self, feed_dict_string_key):
        if self._minimize == None:
            print("set self._minimize operation first!!!")
            sys.exit(1)

        feed_dict = {}
        for key, value in feed_dict_string_key.iteritems():
            feed_dict[self._placeholders[key]] =value

        self._session.run(self._minimize, feed_dict=feed_dict)
        for flush in self._flushShadowNetOps:
            self._session.run(flush)

    def anyNamedOperation(self, name, feed_dict_string_key, isUpdate):
        if self._anyNamedOperations[name] == None:
            print("set the anyoperation you named !!!")
            sys.exit(1)

        feed_dict = {}
        for key, value in feed_dict_string_key.iteritems():
            feed_dict[self._placeholders[key]] =value

        ret = self._session.run(self._anyNamedOperations[name], feed_dict=feed_dict)

        if isUpdate:
            for flush in self._flushShadowNetOps:
                self._session.run(flush)

        return ret


    def getOutValueSmartly(self, feed_dict_string_key):
        if self._ema == None:
            return self.__getOutValue__(feed_dict_string_key)
        else:
            return self.__getShadownOutValue__(feed_dict_string_key)


    def __getShadownOutValue__(self, feed_dict_string_key):
        if self._shadowOut == None:
            print("set out layer first!!!")
            sys.exit(1)



        feed_dict = {}
        for key, value in feed_dict_string_key.iteritems():
            feed_dict[self._placeholders[key]] =value

        return self._session.run(self._shadowOut, feed_dict=feed_dict)

    def __getOutValue__(self, feed_dict_string_key):
        if self._out == None:
            print("set out layer first!!!")
            sys.exit(1)

        feed_dict = {}
        for key, value in feed_dict_string_key.iteritems():
            feed_dict[self._placeholders[key]] =value

        return self._session.run(self._out, feed_dict=feed_dict)






#ema = tf.train.ExponentialMovingAverage(0.999)
#sess = tf.InteractiveSession()
#net = NN(sess, ema)











#!/usr/bin/env python
from __future__ import print_function

import tensorflow as tf
import cv2
import sys
from GameEngine import FlappyBirdGameEngine
import random
import numpy as np
from collections import deque
import os

GAME = 'bird' # the name of the game being played for log files
ACTIONS = 2 # number of valid actions
GAMMA = 0.99 # decay rate of past observations
OBSERVE = 100000. # timesteps to observe before training
EXPLORE = 2000000. # frames over which to anneal epsilon
FINAL_EPSILON = 0.0001 # final value of epsilon
INITIAL_EPSILON = 0.01 # starting value of epsilon lpxz
REPLAY_MEMORY = 50000 # number of previous transitions to remember
BATCH = 32 # size of minibatch
FRAME_PER_ACTION = 1
ORIGIN_LENG = 160
ORIGIN_WIDTH = 160
SAVEDIR_PREFIX = "saved_networks"
MODELNAME = "dqn"
SAVING_PERIOD = 10000

game = FlappyBirdGameEngine()

class QNetwork():
    def __init__(self, iterationID): # create: G, sess, saver, savedDir. load: existing model if any
        self._G = tf.Graph()
        self._Sess = tf.Session(graph=self._G)
        with self._Sess.as_default():
            with self._G.as_default():
                self.createQNetwork()

                tf.global_variables_initializer().run()
                self._saver = tf.train.Saver(tf.global_variables())
                self._savedDir = SAVEDIR_PREFIX + str(iterationID)
                checkpoint = tf.train.get_checkpoint_state(self._savedDir)
                if checkpoint and checkpoint.model_checkpoint_path:
                    self._saver.restore(self._Sess, checkpoint.model_checkpoint_path)


    def weight_variable(self, shape):
        initial = tf.truncated_normal(shape, stddev = 0.01)
        return tf.Variable(initial)

    def bias_variable(self, shape):
        initial = tf.constant(0.01, shape = shape)
        return tf.Variable(initial)

    def conv2d(self, x, W, stride):
        return tf.nn.conv2d(x, W, strides = [1, stride, stride, 1], padding = "SAME")

    def max_pool_2x2(self, x):
        return tf.nn.max_pool(x, ksize = [1, 2, 2, 1], strides = [1, 2, 2, 1], padding = "SAME")

    def getRewardScreen(self, game_state, action):
        x_t1_colored, r_t, terminal = game_state.frame_step(action)
        x_t1 = cv2.cvtColor(cv2.resize(x_t1_colored, (ORIGIN_LENG, ORIGIN_WIDTH)), cv2.COLOR_BGR2GRAY)
        ret, x_t1 = cv2.threshold(x_t1, 1, 255, cv2.THRESH_BINARY)
        return ret, x_t1, terminal


    def createQNetwork(self):
        self._screenInputs = tf.placeholder(tf.float32, [None, ORIGIN_LENG, ORIGIN_WIDTH, 4])
        W1 = tf.Variable(tf.truncated_normal([4, 4, 4, 32], stddev = 0.01)) # lpxz: different from the working version
        B1 = tf.Variable(tf.truncated_normal([32], stddev = 0.01))
        tmp0 = tf.nn.relu(tf.nn.bias_add(self.conv2d(self._screenInputs, W1, 4), B1))
        out1 = self.max_pool_2x2(tmp0)
        W2 = tf.Variable(tf.truncated_normal([2, 2, 32, 64], stddev = 0.01)) # lpxz: different from the working version
        B2 = tf.Variable(tf.truncated_normal([64], stddev = 0.01))
        tmp2 = tf.nn.relu(tf.nn.bias_add(self.conv2d(out1, W2, 2), B2))
        out2 = self.max_pool_2x2(tmp2)
        linearOut = tf.reshape(out2, [-1, 1600])
        W3 = tf.Variable(tf.truncated_normal([1600, 512], stddev = 0.01))
        B3 = tf.Variable(tf.truncated_normal([512], stddev = 0.01))
        tmp = tf.nn.relu(tf.add(tf.matmul(linearOut, W3), B3))
        W4 = tf.Variable(tf.truncated_normal([512, ACTIONS], stddev = 0.01))
        B4 = tf.Variable(tf.truncated_normal([ACTIONS], stddev = 0.01))
        self._qValuesForActions = tf.add(tf.matmul(tmp, W4), B4)

        # train:
        self._actionChoiceInputs = tf.placeholder(tf.float32, [None, ACTIONS])
        self._targetValueInputs = tf.placeholder(tf.float32, [None])

        #tf.sub does not exist, simply use "-"
        vvv = tf.multiply(self._actionChoiceInputs, self._qValuesForActions)
        self._observedValues = tf.reduce_sum(vvv, reduction_indices = 1)
        self._cost = tf.reduce_mean(tf.square(self._targetValueInputs - self._observedValues), 0)
        self._optimizer = tf.train.AdamOptimizer(learning_rate=1e-6).minimize(self._cost)


    def makeImageBlock(self,image):
        return np.stack((image, image, image, image), axis=2)

    def updateImageBlock(self,block, image):
        return np.append(block[:, :, 1:], image, axis=2)

    def control(self, referenceNet=None):
        action = np.zeros([ACTIONS])
        action[0] = 1
        game_state = game.initialState()
        reward, x_t1, _ = game.step(action, game_state)
        block = self.makeImageBlock(x_t1)

        replayBuffer = deque()
        episilon = INITIAL_EPSILON
        counter = 0
        while True:
            counter += 1
            action = np.zeros([ACTIONS])

            # read -> episilon greedy control
            if random.random() < episilon:
                randIndex = random.randint(0, ACTIONS-1)
                action[randIndex] = 1
                qvaluesForActions = None # do not inherit from last timestamp
            else:
                with self._Sess.as_default():
                    with self._G.as_default():
                        qvaluesForActions = self._qValuesForActions.eval(feed_dict={self._screenInputs: [block]})[0]
                        index = np.argmax(qvaluesForActions)
                        action[index] = 1

            # scale down epsilon
            if episilon > FINAL_EPSILON and counter > OBSERVE:
                episilon -= (INITIAL_EPSILON - FINAL_EPSILON) / EXPLORE

            reward, x_t1, terminal = game.step(action, game_state)

            x_t1 = np.reshape(x_t1, (ORIGIN_LENG, ORIGIN_WIDTH, 1))
            blockNew = self.updateImageBlock(block, x_t1)

            # feedback -> replay buffer
            if len(replayBuffer) >= REPLAY_MEMORY:
                replayBuffer.popleft()

            replayBuffer.append((block, action, reward, terminal, blockNew))
            if counter >= OBSERVE:
                # train
                batchObservations = random.sample(replayBuffer, BATCH)
                screenInputVals = [x[0] for x in batchObservations]
                actionChoiceInputVals = [x[1] for x in batchObservations]
                rewards = [x[2] for x in batchObservations]
                terminals = [x[3] for x in batchObservations]
                blockNews = [x[4] for x in batchObservations]
                targetValueInputVals = np.zeros([BATCH])
                for i in range(BATCH):
                    if terminals[i]:
                        targetValueInputVals[i] = rewards[i]
                    else:
                        # target value.
                        if referenceNet == None: # from self
                            with self._Sess.as_default():
                                with self._G.as_default():
                                    qvaluesForNewState = self._qValuesForActions.eval(feed_dict={self._screenInputs: [blockNews[i]]})[0]
                                    targetValueInputVals[i] = rewards[i] + GAMMA * np.max(qvaluesForNewState, axis=0)
                        else: # from reference net
                            with referenceNet._Sess.as_default():
                                with referenceNet._G.as_default():
                                    qvaluesForNewState = referenceNet._qValuesForActions.eval(feed_dict={referenceNet._screenInputs: [blockNews[i]]})[0]
                                    targetValueInputVals[i] = rewards[i] + GAMMA * np.max(qvaluesForNewState, axis=0)


                with self._Sess.as_default():
                    with self._G.as_default():
                        self._optimizer.run(feed_dict={self._screenInputs: screenInputVals, self._actionChoiceInputs: actionChoiceInputVals, self._targetValueInputs:targetValueInputVals})

            if qvaluesForActions == None:
                print("TIMESTEP", counter,  "/ ACTION", np.argmax(action), "/ REWARD", reward, "/ terminal ",  terminal)
            else:
                print("TIMESTEP", counter,  "/ ACTION", np.argmax(action), "/ REWARD", reward, "/ terminal ",  terminal, "/ Q_MAX ", np.max(qvaluesForActions))

            block = blockNew


            if counter % SAVING_PERIOD == 0:
                if not os.path.exists(self._savedDir):
                    os.makedirs(self._savedDir)
                self._saver.save(self._Sess, self._savedDir + '/' + MODELNAME) # overwrite the last iteration







# iteartively refine the representation of the images.
# the network trained in last iteration using the coarser images are used to guide the current iteartion



def main():
    qNetOld = QNetwork(iterationID=0)
    qNet = QNetwork(iterationID=1)
    qNet.control(referenceNet=qNetOld)



if __name__ == "__main__":
    main()



from NN import NN
import tensorflow as tf

class CNN(NN):
    def __init__(self, session, hasShadowNet=True):
        NN.__init__(self, session, hasShadowNet)

    def conv2d(self, x, filterWindow, strides):
        return tf.nn.conv2d(x, filterWindow, strides = strides, padding = "SAME")

    def max_pool(self, x, filterWindow, strides): # we provided the default parameter
        return tf.nn.max_pool(x, ksize = filterWindow, strides = strides, padding = "SAME")


#        tmp0 = tf.nn.relu(tf.nn.bias_add(self.conv2d(self._screenInputs, W1, 4), B1))
    def buildConvReluWire(self, layer, filterWindow, strides): # strides = [1,2,2,1]
        # basic
        W1 = self.weight_variable(shape = filterWindow)
        B1 = self.bias_variable(shape=[filterWindow[-1]])
        tmp = tf.nn.bias_add(self.conv2d(layer, W1, strides), B1)
        outputLayer = tf.nn.relu(tmp)
        weightName = self.uniqueWeightName()
        self._variables[weightName] = W1 # register the created variables
        biasName = self.uniqueBiasName()
        self._variables[biasName] = B1
        layerName = self.uniqueLayerName()
        self._layers[layerName] = outputLayer # register the layer

        # visualization
        label = weightName+":"+ self.shape2string(filterWindow) + "," + biasName + ":"+self.shape2string([filterWindow[-1]])
        self.createGraphvizEdge(self.lookupLayerName(layer), layerName, label)

        # shadow part:
        if self._ema != None:
            flush = self._ema.apply([W1, B1]) # create some symbolic representation of shadow, otherwise, the following "average" call would return None
            self._flushShadowNetOps.append(flush)
            W1shadow = self._ema.average(W1)
            B1shadow = self._ema.average(B1)
            layerShadow = self._layers2shadow[layer]
            tmpshadow = tf.nn.bias_add(self.conv2d(layerShadow, W1shadow, strides), B1shadow)
            outputLayershadow = tf.nn.relu(tmpshadow)
            self._variables2shadow[W1] = W1shadow
            self._variables2shadow[B1] = B1shadow
            self._layers2shadow[outputLayer] = outputLayershadow
            self.createGraphvizEdge(self.shadowPrefixed(self.lookupLayerName(layer)), self.shadowPrefixed(layerName), label)


        return outputLayer

    def buildConvWire(self, layer, filterWindow, strides): # strides = [1,2,2,1]
        # basic
        W1 = self.weight_variable(shape = filterWindow)
        B1 = self.bias_variable(shape=[filterWindow[-1]])
        outputLayer = tf.nn.bias_add(self.conv2d(layer, W1, strides), B1)
        weightName = self.uniqueWeightName()
        self._variables[weightName] = W1 # register the created variables
        biasName = self.uniqueBiasName()
        self._variables[biasName] = B1
        layerName = self.uniqueLayerName()
        self._layers[layerName] = outputLayer # register the layer

        # visualization
        label = weightName+":"+ self.shape2string(filterWindow) + "," + biasName + ":"+self.shape2string([filterWindow[-1]])
        self.createGraphvizEdge(self.lookupLayerName(layer), layerName, label)

        # shadow part:
        if self._ema != None:
            flush = self._ema.apply([W1, B1]) # create some symbolic representation of shadow, otherwise, the following "average" call would return None
            self._flushShadowNetOps.append(flush)
            W1shadow = self._ema.average(W1)
            B1shadow = self._ema.average(B1)
            layerShadow = self._layers2shadow[layer]
            outputLayershadow = tf.nn.bias_add(self.conv2d(layerShadow, W1shadow, strides), B1shadow)
            self._variables2shadow[W1] = W1shadow
            self._variables2shadow[B1] = B1shadow
            self._layers2shadow[outputLayer] = outputLayershadow
            self.createGraphvizEdge(self.shadowPrefixed(self.lookupLayerName(layer)), self.shadowPrefixed(layerName), label)


        return outputLayer


    def buildMaxpoolWire(self, layer, filterWindow=[1, 2, 2, 1], strides=[1, 2, 2, 1]): # strides = [1,2,2,1]
        # basic
        outputLayer = self.max_pool(layer, filterWindow, strides)
        layerName = self.uniqueLayerName()
        self._layers[layerName] = outputLayer # register the layer

        # visualization
        label = "maxpool2*2"
        self.createGraphvizEdge(self.lookupLayerName(layer), layerName, label)

        # shadow part:
        if self._ema != None:
            layerShadow = self._layers2shadow[layer]
            outputLayershadow = self.max_pool(layerShadow, filterWindow, strides)
            self._layers2shadow[outputLayer] = outputLayershadow
            self.createGraphvizEdge(self.shadowPrefixed(self.lookupLayerName(layer)), self.shadowPrefixed(layerName), label)

        return outputLayer

    def buildFlattenWire(self, layer, flattenAs): # strides = [1,2,2,1]
        # basic
        outputLayer = tf.reshape(layer, flattenAs)
        layerName = self.uniqueLayerName()
        self._layers[layerName] = outputLayer # register the layer

        # visualization
        label = "flattenAs"+self.shape2string(flattenAs)
        self.createGraphvizEdge(self.lookupLayerName(layer), layerName, label)

        # shadow part:
        if self._ema != None:
            layerShadow = self._layers2shadow[layer]
            outputLayershadow = tf.reshape(layerShadow, flattenAs)
            self._layers2shadow[outputLayer] = outputLayershadow
            self.createGraphvizEdge(self.shadowPrefixed(self.lookupLayerName(layer)), self.shadowPrefixed(layerName), label)

        return outputLayer



import tensorflow as tf
from NN import NN
class Critic(NN):
    def __init__(self, session, hasShadowNet, state_size, action_size, hidden_state_size):
        NN.__init__(self, session, hasShadowNet)
        # nothing special:
        inputStates = self.buildInputLayer("inputStates", shape=[None, state_size])
        inputActions = self.buildInputLayer("inputActions", shape=[None, action_size])
        inputYs = self.buildInputLayer("inputYs", shape=[None])
        h1 = self.buildLinearReluWire(inputStates, [state_size, hidden_state_size])
        #for i in range(numOfHiddenLayers-1): # repeat (numOfHiddenLayers-1) times
        h1 = self.buildLinearReluWire(h1, [hidden_state_size, hidden_state_size])
        h1 = self.buildJointLinearReluWire(h1, [hidden_state_size, hidden_state_size], inputActions, [action_size, hidden_state_size])
        tmp1 = self.buildLinearWire(h1, [hidden_state_size, action_size])
        out = self.buildReduceSum(tmp1, reduction_indices=1)
        self.setOutLayer(out)

        # a bit unique:
        # action_size =1, so we do not need reduce_sum or action_choice
        self.error = tf.reduce_mean(tf.square(inputYs - out)) # all, every row has only one, let us take fast path
        self.addMinimizeOperation(tf.train.AdamOptimizer(0.001).minimize(self.error))
        self.addAnyNamedOperation("goa", tf.gradients(out, inputActions))





#from distutils.core import setup
from setuptools import setup, find_packages
setup(
  name = 'rlsimple',
  packages = find_packages(),
    #packages = ['rlsimple', 'rlsimple.DDPG', 'rlsimple.DQN', 'rlsimple.FlappyBirdEnv'], # this must be the same as the name above
  version = '1.1',
  description = 'reinforcement learning simple library',
  author = 'Rabbit Noname',
  author_email = 'faithoptimistic@gmail.com',
  url = 'https://github.com/rabbitnoname/rlsimple', # use the URL to the github repo
  download_url = 'https://github.com/rabbitnoname/rlsimple/archive/master.zip', # I'll explain this in a second
  keywords = ['reinforcement learning', 'DDPG', 'DQN', 'Actor Critic', 'simple library'], # arbitrary keywords
  classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 2.7',
    ]
)



import sys
import os

currentDir = os.path.dirname(os.path.abspath(__file__))
parentDir = os.path.abspath(os.path.join(currentDir,os.pardir)) # this will return parent directory.
sys.path.insert(0,parentDir)

import numpy as np
import tensorflow as tf


from Actor import Actor
from Critic import Critic
from ReplayBuffer import ReplayBuffer
from GameEngine import OpenAIGameEngine
# REPLAY BUFFER CONSTS
BUFFER_SIZE = 10000
BATCH_SIZE = 128
# FUTURE REWARD DECAY
GAMMA = 0.99

ENVIRONMENT_NAME = 'Pendulum-v0'
env = OpenAIGameEngine(ENVIRONMENT_NAME)



sess = tf.InteractiveSession()
state_size, action_size, action_high, action_low = env.gameParameters()
actor = Actor(sess, True, state_size, action_size, 200)
actor._session.run(tf.initialize_all_variables())
critic = Critic(sess, True, state_size, action_size, 300)
critic._session.run(tf.initialize_all_variables())
buffer = ReplayBuffer(BUFFER_SIZE)


def task():
    for ep in range(10000):
        s = env.initialState()
        Totoal = 0
        # what if the action is beyond the scope?
        for iteration  in range(100):
            # select the action with actor model.
            env.show()
            a = actor.getOutValueSmartly({"inputStates": [s]})[0] + (np.random.randn(1) / (ep + iteration + 1))
            r, s1, terminated = env.step(a)
            Totoal += r
            buffer.add(s, a, r, s1, terminated) #state, action, reward, new_state, done

            # update critic
            batch = buffer.getBatch(batch_size=BATCH_SIZE)
            S_ = [e[0] for e in batch]
            A_ = [e[1] for e in batch]
            R_ = [e[2] for e in batch]
            S1_ = [e[3] for e in batch]
            notTerminated = [1.-e[4] for e in batch]
            A1_ = actor.getOutValueSmartly({"inputStates":S1_})
            Q_S1_A1_ = critic.getOutValueSmartly({"inputStates":S1_, "inputActions":A1_})
            Ys = R_ + GAMMA * Q_S1_A1_ * notTerminated
            critic.minimizeSmartly({"inputStates":S_, "inputActions":A_, "inputYs":Ys})
            # update actor
            A = actor.getOutValueSmartly({"inputStates":S_})
            Qgradients = critic.anyNamedOperation("goa", {"inputStates":S_, "inputActions":A_}, False)[0]
            actor.ascentSmartly({"inputStates":S_, "Qgradients":Qgradients})

            s = s1

        print "EPISODE ", ep, "ENDED UP WITH REWARD: ", Totoal

# monitored run
env.monitor(task)



#!/usr/bin/env python
from __future__ import print_function
import sys
import os

currentDir = os.path.dirname(os.path.abspath(__file__))
parentDir = os.path.abspath(os.path.join(currentDir,os.pardir)) # this will return parent directory.
sys.path.insert(0,parentDir)

from CNN import CNN
import tensorflow as tf
import cv2
from GameEngine import FlappyBirdGameEngine
import random
import numpy as np
from collections import deque

GAME = 'bird' # the name of the game being played for log files
ACTIONS = 2 # number of valid actions
GAMMA = 0.99 # decay rate of past observations
OBSERVE = 100000. # timesteps to observe before training
EXPLORE = 2000000. # frames over which to anneal epsilon
FINAL_EPSILON = 0.0001 # final value of epsilon
INITIAL_EPSILON = 0.01 # starting value of epsilon lpxz
REPLAY_MEMORY = 50000 # number of previous transitions to remember
BATCH = 32 # size of minibatch
FRAME_PER_ACTION = 1
ORIGIN_LENG = 160
ORIGIN_WIDTH = 160
SAVEDIR_PREFIX = "saved_networks"
MODELNAME = "dqn"
SAVING_PERIOD = 10000

game = FlappyBirdGameEngine()

class QNetwork(CNN):
    def __init__(self, session, hasShadowNet): # create: G, sess, saver, savedDir. load: existing model if any
        CNN.__init__(self, session, hasShadowNet)
        self._screenInputs = self.buildInputLayer("_screenInputs", shape=[None, ORIGIN_LENG, ORIGIN_WIDTH, 4])
        tmp0 = self.buildConvReluWire(self._screenInputs, [4, 4, 4, 32], [1,4,4,1]) # stride=[1,4,4,1], where 4,4 are important
        out1 = self.buildMaxpoolWire(tmp0)
        tmp2 = self.buildConvReluWire(out1, [2, 2, 32, 64], [1,2,2,1]) # stride=[1,4,4,1], where 4,4 are important
        out2 = self.buildMaxpoolWire(tmp2)
        linearOut = self.buildFlattenWire(out2, [-1, 1600])
        tmp = self.buildLinearReluWire(linearOut, [1600, 512])
        self._qValuesForActions = self.buildLinearWire(tmp, [512, ACTIONS])
        self.setOutLayer(self._qValuesForActions)

        # train:
        #self._actionChoiceInputs = tf.placeholder(tf.float32, [None, ACTIONS])
        self._actionChoiceInputs = self.buildInputLayer("_actionChoiceInputs", shape=[None, ACTIONS])
        #self._targetValueInputs = tf.placeholder(tf.float32, [None])
        self._targetValueInputs = self.buildInputLayer("_targetValueInputs", shape=[None])

        #tf.sub does not exist, simply use "-"
        vvv = tf.multiply(self._actionChoiceInputs, self._qValuesForActions)
        self._observedValues = tf.reduce_sum(vvv, reduction_indices = 1)
        self._cost = tf.reduce_mean(tf.square(self._targetValueInputs - self._observedValues), 0)
        self._optimizer = tf.train.AdamOptimizer(learning_rate=1e-6).minimize(self._cost)
        self.addMinimizeOperation(self._optimizer)
        # self._G = tf.Graph()
        # self._Sess = tf.Session(graph=self._G)
        # with self._Sess.as_default():
        #     with self._G.as_default():
        #         self.createQNetwork()
        #
        #         tf.global_variables_initializer().run()
        #         self._saver = tf.train.Saver(tf.global_variables())
        #         self._savedDir = SAVEDIR_PREFIX
        #         checkpoint = tf.train.get_checkpoint_state(self._savedDir)
        #         if checkpoint and checkpoint.model_checkpoint_path:
        #             self._saver.restore(self._Sess, checkpoint.model_checkpoint_path)



    def getRewardScreen(self, game_state, action):
        x_t1_colored, r_t, terminal = game_state.frame_step(action)
        x_t1 = cv2.cvtColor(cv2.resize(x_t1_colored, (ORIGIN_LENG, ORIGIN_WIDTH)), cv2.COLOR_BGR2GRAY)
        ret, x_t1 = cv2.threshold(x_t1, 1, 255, cv2.THRESH_BINARY)
        return ret, x_t1, terminal




    def makeImageBlock(self,image):
        return np.stack((image, image, image, image), axis=2)

    def updateImageBlock(self,block, image):
        return np.append(block[:, :, 1:], image, axis=2)

    def control(self, referenceNet=None):
        action = np.zeros([ACTIONS])
        action[0] = 1
        game_state = game.initialState()
        reward, x_t1, _ = game.step(action, game_state)
        block = self.makeImageBlock(x_t1)

        replayBuffer = deque()
        episilon = INITIAL_EPSILON
        counter = 0
        while True:
            counter += 1
            action = np.zeros([ACTIONS])

            # read -> episilon greedy control
            if random.random() < episilon:
                randIndex = random.randint(0, ACTIONS-1)
                action[randIndex] = 1
                qvaluesForActions = None # do not inherit from last timestamp
            else:
                qvaluesForActions = self._qValuesForActions.eval(feed_dict={self._screenInputs: [block]})[0]
                index = np.argmax(qvaluesForActions)
                action[index] = 1

            # scale down epsilon
            if episilon > FINAL_EPSILON and counter > OBSERVE:
                episilon -= (INITIAL_EPSILON - FINAL_EPSILON) / EXPLORE

            reward, x_t1, terminal = game.step(action, game_state)

            x_t1 = np.reshape(x_t1, (ORIGIN_LENG, ORIGIN_WIDTH, 1))
            blockNew = self.updateImageBlock(block, x_t1)

            # feedback -> replay buffer
            if len(replayBuffer) >= REPLAY_MEMORY:
                replayBuffer.popleft()

            replayBuffer.append((block, action, reward, terminal, blockNew))
            if counter >= OBSERVE:
                # train
                batchObservations = random.sample(replayBuffer, BATCH)
                screenInputVals = [x[0] for x in batchObservations]
                actionChoiceInputVals = [x[1] for x in batchObservations]
                rewards = [x[2] for x in batchObservations]
                terminals = [x[3] for x in batchObservations]
                blockNews = [x[4] for x in batchObservations]
                targetValueInputVals = np.zeros([BATCH])
                for i in range(BATCH):
                    if terminals[i]:
                        targetValueInputVals[i] = rewards[i]
                    else:
                        # target value.
                        if referenceNet == None: # from self
                            #qvaluesForNewState = self._qValuesForActions.eval(feed_dict={self._screenInputs: [blockNews[i]]})[0]
                            qvaluesForNewState = self.getOutValueSmartly({"_screenInputs": [blockNews[i]]})[0]
                            targetValueInputVals[i] = rewards[i] + GAMMA * np.max(qvaluesForNewState, axis=0)
                        # else: # from reference net
                        #     qvaluesForNewState = referenceNet._qValuesForActions.eval(feed_dict={referenceNet._screenInputs: [blockNews[i]]})[0]
                        #     targetValueInputVals[i] = rewards[i] + GAMMA * np.max(qvaluesForNewState, axis=0)


                self.minimizeSmartly({"_screenInputs": screenInputVals, "_actionChoiceInputs": actionChoiceInputVals, "_targetValueInputs":targetValueInputVals})
                #self._optimizer.run(feed_dict={self._screenInputs: screenInputVals, self._actionChoiceInputs: actionChoiceInputVals, self._targetValueInputs:targetValueInputVals})

            if qvaluesForActions == None:
                print("TIMESTEP", counter,  "/ ACTION", np.argmax(action), "/ REWARD", reward, "/ terminal ",  terminal)
            else:
                print("TIMESTEP", counter,  "/ ACTION", np.argmax(action), "/ REWARD", reward, "/ terminal ",  terminal, "/ Q_MAX ", np.max(qvaluesForActions))

            block = blockNew


            # if counter % SAVING_PERIOD == 0:
            #     if not os.path.exists(self._savedDir):
            #         os.makedirs(self._savedDir)
            #     self._saver.save(self._Sess, self._savedDir + '/' + MODELNAME) # overwrite the last iteration







# iteartively refine the representation of the images.
# the network trained in last iteration using the coarser images are used to guide the current iteartion



def main():
    sess = tf.InteractiveSession()
    qNet = QNetwork(sess, True)
    tf.global_variables_initializer().run()

    qNet.control()



if __name__ == "__main__":
    main()
